/****************************************************************************
*
*	  Header Filename:  Action_Request.h
*	  Module Name:      Action_Request
*	  Description:      Specifies and controls a particular action required
*                       as part of a simulation Scenario
*	  Author:           Chris Gordon-Smith
*
*     Copyright (C) 2005, 2006, 2008 Chris Gordon-Smith
*     Email: See http://www.simsoup.info/
*
*     This program is free software; you can redistribute it and/or modify it
*     under the terms of the GNU General Public License Version 2, as published 
*     by the Free Software Foundation.
*
*     This program is distributed in the hope that it will be useful, but
*     WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
*     or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License
*     for more details.
*
*     You should have received a copy of the GNU General Public License along
*     with this program; if not, write to the Free Software Foundation, Inc.,
*     59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
*
*     See the file Licence.txt for a copy of the GNU General Public License.
*
*      Amendment History:- 
*
*              Version	  Date                        Update Description
*              =======    ===========                 ==================
*
*
*
*
****************************************************************************/

#ifndef Action_Request_H
#define Action_Request_H

class   Molecule_Type;
class   Add_Constr_ReqRqstr;
class	Add_Fission_ReqRqstr;
class	Add_Transfm_ReqRqstr;
class   Add_Cat_Trans_ReqRqstr;
class   Add_Maint_MolType_Spec_ReqRqstr;
class   Add_Molecule_ReqRqstr;
class   Add_MolType_ReqRqstr;
class   ChemEvol_ReqRqstr;
class	Empty_Reactor_ReqRqstr;
class   Remove_Fission_ReqRqstr;
class   Remove_Construction_ReqRqstr;
class   Remove_Transformation_ReqRqstr;
class	Remove_Maint_MolType_Spec_ReqRqstr;
class	Set_Reactor_Vol_ReqRqstr;
class   Set_Reactor_Volume_ReqRqstr;
class	Set_Max_Rand_Rate_Const_Constr_ReqRqstr;
class   Set_Max_Rand_Rate_Const_Fission_ReqRqstr;
class   Set_Max_Rate_Const_Transformation_ReqRqstr;
class   Set_Rate_Const_Model_ReqRqstr;
class   Terminate_Scenario_ReqRqstr;
class   MolType_Spec; 
class   Mass_Spec;
class   PotEnergy_Spec;
class   ACE_Spec;

class   Interaction_Type;
class   Realisable_Interaction_Type;


class   Action_Request

//  Base Action Request class

{
    public: 
        Action_Request(int Time, int Repeat_Interval);
        virtual ~Action_Request();

        string  Get_Act_Rslt_Tx() const;
        int  Get_Time() const;
        int  Get_Repeat_Interval() const;
        bool Get_Repeat_Req_Fl() const;
        bool Get_Fail_Fl() const;
        bool Get_Prev_Fail_Fl() const ;
        bool Setup_Repeat_If_Reqd();

        virtual void Execute() = 0;
        virtual string Get_Type() const = 0;
        virtual string Get_Subject() const = 0;
        virtual void Get_Params(list<string>& Param_List) const = 0;

        static bool Compare_Time(const Action_Request* x, const Action_Request* y);

    protected:
        int  Time_;                 // Execution time
        bool Repeat_Req_;           // Set true if this is an automatically generated repeat Action Request
        int  Repeat_Interval_;      // 0 means no repeat
        bool Fail_Fl_;              // true means last execution failed
        bool Prev_Exec_Fail_Fl_;    // Set true if this is a Repeat Request 
                                    // and any Execution attempt in the series has failed
        string Fail_Reason_Tx_;                 // Reason for failure of execution                                       
        string Success_Qualifier_Tx_;           // Qualifier success. 
                                                // Only used if there are various success outcomes
        static  const string Repeat_Req_Tx;     // For logging Action Requests that are to be repeated

        virtual string Succ_Msg() const = 0 ;
        virtual string Fail_Msg() const = 0 ;  // Derived classes must provide, but if Execute for
                                               // the derived class cannot fail, then 
                                               // Fail_Msg should never be called and should Assert(false)

        void Exec_Success(const char*  Tx_Pt);
        void Exec_Fail(const char*  Tx_Pt);
        void Exec_Fail(const string*  Tx_Pt);

    private:
        Action_Request(const Action_Request&);  // Disallow copying
        Action_Request& operator= (const Action_Request&);  // Disallow assignment
};
//----------------------------------------------------------------------------------------------------------------

class   ActReq_Help

//  Action_Request Helper Functiona

{
    public:
        static  bool MolType_Spec_2_Id(const MolType_Spec* MT_Spec_Pt, 
                                       int& MT_Id, 
                                       string& Rslt_Tx);
//        static  string Get_Equals_Tx()  {return string(" = ");}
//        static  string Get_Warn_Tx() {return string(" **** WARNING **** ");}
};
//----------------------------------------------------------------------------------------------------------------
class   Add_Cat_Trans_Req: public Action_Request

//  Action Request to add a simple Catalytic Transformation

{
    public:
        Add_Cat_Trans_Req(Add_Cat_Trans_ReqRqstr*  Requestor_Ptr);
        string Execute(Action_Request*& Repeat_Act_Req_Ptr); 
        int    Get_Number();
        string Get_Type();
        string Get_Subject();
        void  Get_Params(list<string>& Param_List) const;
        
    private:
        Add_Cat_Trans_Req(const Add_Cat_Trans_Req&);
        int     Num_To_Add;
        int     Sub_Net_Id;
        string  Reactant_Spec;
        string  Catalyst_Spec;
        string  Product_Spec;

        static int  Specify_Cat_Trans(bool Cat_Trans_Reactant_Random_Flag, int&     Cat_Trans_Reactant_Index,
                                      bool Cat_Trans_Product_Random_Flag,  int&     Cat_Trans_Product_Index,
                                      bool Cat_Trans_Catalyst_Random_Flag, int&     Cat_Trans_Catalyst_Index,
                                      int&     Intermediate_Product_Index,   //  Compound Interaction Intermediate 
				                      double&  Construction_Actv_Cplx_Energy,
				                      double&  Fission_Actv_Cplx_Energy);
        void    Make_And_Notify_Int_Types_For_Cat_Trans();

};
//----------------------------------------------------------------------------------------------------------------

class   Add_Constr_Req: public Action_Request

//  Action Request to add a Construction

{
    public:
  		explicit Add_Constr_Req(Add_Constr_ReqRqstr*  Reqstr_Ptr);
        ~Add_Constr_Req();
        void   Execute(); 
        int    Get_Number()  const;        
        string Get_Type()  const;
        string Get_Subject()  const;
        void  Get_Params(list<string>& Param_List) const;
        
    private:
        virtual string  Succ_Msg() const;
        virtual string  Fail_Msg() const;
        const int   Num_To_Add_;
        const MolType_Spec*  R1_Spec_Pt_;
        const MolType_Spec*  R2_Spec_Pt_;
        const MolType_Spec*  P_Spec_Pt_;
        const ACE_Spec*  ACE_Spec_Pt_;
};
//-----------------------------------------------------------------------------------------------------------------

class   Add_Fission_Req: public Action_Request

//  Action Request to add a Fission

{
    public:
        explicit Add_Fission_Req(Add_Fission_ReqRqstr*  Reqstr_Ptr);
        ~Add_Fission_Req();
        void   Execute(); 
        int    Get_Number()  const;        
        string Get_Type()  const;
        string Get_Subject()  const;
        void  Get_Params(list<string>& Param_List) const;
        
    private:
        virtual string  Succ_Msg() const;
        virtual string  Fail_Msg() const;
        const int   Num_To_Add_;
        const MolType_Spec*  R_Spec_Pt_;
        const MolType_Spec*  P1_Spec_Pt_;
        const MolType_Spec*  P2_Spec_Pt_;
        const ACE_Spec*  ACE_Spec_Pt_;
};
//-----------------------------------------------------------------------------------------------------------------

class   Add_Maint_MolType_Spec_Req: public Action_Request

//  Action Request to add a Maintained Molecule Type Specifier

{
    public:
        Add_Maint_MolType_Spec_Req(Add_Maint_MolType_Spec_ReqRqstr*  Requestor_Ptr);
//        Add_Maint_MolType_Spec_Req(int Action_Time, 
//                                   int Num_To_Add, 
//                                   int Num_Mols_Fixd,
//                                   int Repeat_Interval, 
//                                   bool Repeat_Req);

        string Execute(Action_Request*& Repeat_Act_Req_Ptr); 
        int    Get_Number() const;
        string Get_Type() const;
        string Get_Subject() const;
        void  Get_Params(list<string>& Param_List) const;
        
    private:
        int    Num_To_Add;   // For Fixed Inflow
        int    Num_Mols_Fixed;     // For Fixed number of Molecules
        double Tracker_Probability;
        int    Molecule_Type_Index;
};
//-----------------------------------------------------------------------------------------------------------------

class   Add_Molecule_Req: public Action_Request

//  Action Request to add a Molecule

{
    public:
        explicit Add_Molecule_Req(const Add_Molecule_ReqRqstr*  Requestor_Ptr);
        ~Add_Molecule_Req();
        void    Execute(); 
        int     Get_Number() const;
        string  Get_Type() const;
        string  Get_Subject() const;
        void    Get_Params(list<string>& Param_List) const;
        
    private:
//        enum    Succ_Code {Rnd_MolType_Cd,       // Random MolType created
//                           Fixed_MolType_Cd}     // Fixed MolType created             
//                Succ_Cd_;
  //
//        enum    Fail_Code    {MolType_NotExist_Cd}  // Failed due to MolType not existing
//                Fail_Cd_;

        MolType_Spec*   MolType_Spec_Pt_;
        int     Num_To_Add_;

        virtual string  Succ_Msg() const;
        virtual string  Fail_Msg() const;
        
        Add_Molecule_Req(const Add_Molecule_Req&);
        Add_Molecule_Req& operator= (const Add_Molecule_Req&);  
};
//-----------------------------------------------------------------------------------------------------------------

class   Add_MolType_Req: public Action_Request

//  Action Request to add a Molecule Type

{
    public:
	    explicit Add_MolType_Req(const Add_MolType_ReqRqstr*  Requestor_Ptr);
        ~Add_MolType_Req();
        void   Execute(); 
        int    Get_Number() const;
        string Get_Type() const;
        string Get_Subject() const;
        void  Get_Params(list<string>& Param_List) const;
        
    private:

        string  Succ_Msg() const;
        string  Fail_Msg() const
            {Assert(false, "Add_MolType_Req::Fail_Msg - Should never be called");
             return string ("");}

        int       Num_To_Add_;
        Mass_Spec*          Mass_Spec_Pt_;
        PotEnergy_Spec*    PotEnergy_Spec_Pt_;

        Add_MolType_Req(const Add_MolType_Req&);
        Add_MolType_Req& operator= (const Add_MolType_Req&);  
};
//----------------------------------------------------------------------------------------------------------------

class   ChemEvol_Req: public Action_Request

//  Action Request for a Chemistry Evolution step

{
    public:
        explicit ChemEvol_Req(const ChemEvol_ReqRqstr*  Reqstr_Ptr);
        ~ChemEvol_Req();
        void   Execute(); 
        int    Get_Number()  const;        
        string Get_Type()  const;
        string Get_Subject()  const;
        void  Get_Params(list<string>& Param_List) const;
        
    private:
        double Last_Fitness_;   // Heat Release Rate before execution of the Evolution Step that preceeded
                                // this one. This will be used to determine whether the last step
                                // was successful, and therefore to determine whether the change
                                // it introduced should be kept or discarded.

        struct  Mutation_Specifier
        {
            Interaction_Type*  Last_IntType_Added_Pt;
            Realisable_Interaction_Type*  Last_R_Int_Type_Disabled_Pt;
            Interaction_Type*  Last_IntType_Disabled_Pt;
        };
        Mutation_Specifier  Last_Mutation_;
        Mutation_Specifier  Current_Mutation_;

        virtual string  Succ_Msg() const;
        virtual string  Fail_Msg() const;
};
//----------------------------------------------------------------------------------------------------------------

class   Remove_Fission_Req: public Action_Request

//  Action Request to remove a Fission

{
    public:
		Remove_Fission_Req(Remove_Fission_ReqRqstr*  Requestor_Ptr);
        string Execute(Action_Request*& Repeat_Act_Req_Ptr); 
        int    Get_Number();        
        string Get_Type();
        string Get_Subject();
        void  Get_Params(list<string>& Param_List) const;
        
    private:
        int     Num_To_Remove;
        string  Reactant_Spec;
        string  Product1_Spec;
        string  Product2_Spec;
};
//----------------------------------------------------------------------------------------------------------------

class   Remove_Construction_Req: public Action_Request

//  Action Request to remove a Construction

{
    public:
        Remove_Construction_Req(Remove_Construction_ReqRqstr*  Requestor_Ptr);
        string Execute(Action_Request*& Repeat_Act_Req_Ptr); 
        int    Get_Number();        
        string Get_Type();
        string Get_Subject();
        void  Get_Params(list<string>& Param_List) const;
        
    private:
        int     Num_To_Remove;
        string  Reactant1_Spec;
        string  Reactant2_Spec;
        string  Product_Spec;
};
//----------------------------------------------------------------------------------------------------------------

class   Add_Transformation_Req: public Action_Request

//  Action Request to add a Transformation

{
    public:
        explicit Add_Transformation_Req(Add_Transfm_ReqRqstr*  Reqstr_Ptr);
        ~Add_Transformation_Req();
        void   Execute(); 
        int    Get_Number()  const;        
        string Get_Type()  const;
        string Get_Subject()  const;
        void  Get_Params(list<string>& Param_List) const;
        
    private:
        virtual string  Succ_Msg() const;
        virtual string  Fail_Msg() const;
        const int   Num_To_Add_;
        const MolType_Spec*  R_Spec_Pt_;
        const MolType_Spec*  P_Spec_Pt_;
        const ACE_Spec*  ACE_Spec_Pt_;
};
//-----------------------------------------------------------------------------------------------------------------


class   Empty_Reactor_Req: public Action_Request

//  Action Request to empty the Reactor

{
    public:
        Empty_Reactor_Req(Empty_Reactor_ReqRqstr*  Requestor_Ptr);
        string Execute(Action_Request*& Repeat_Act_Req_Ptr); 
        int    Get_Number();        
        string Get_Type();
        string Get_Subject();
        void  Get_Params(list<string>& Param_List) const;
        
    private:

};
//----------------------------------------------------------------------------------------------------------------
/*
class   Evolution_Step_Req: public Action_Request

//  Action Request to execute an Evolution Step

{
      

    public:
//        struct  Mutation_Specifier
//        {
//            Interaction_Type*  Last_IntType_Added_Ptr;
//            Realisable_Interaction_Type*  Last_R_Int_Type_Disabled_Ptr;
//            Interaction_Type*  Last_IntType_Disabled_Ptr;
//        };
        Evolution_Step_Req(Evolution_Step_ReqRqstr*  Requestor_Ptr);
        string Execute(Action_Request*& Repeat_Act_Req_Ptr); 
        int    Get_Number();        
        string Get_Type();
        string Get_Subject();
        void  Get_Params(list<string>& Param_List) const;
        
    private:
        double Last_Fitness;          // Heat Release Rate before execution of the Evolution Step that preceeded
                                      // this one. This will be used to determine whether the last step
                                      // was successful, and therefore to determine whether the change 
                                      // it introduced should be kept or discarded.
        
};*/
//-----------------------------------------------------------------------------------------------------------------

class   Remove_Maint_MolType_Spec_Req: public Action_Request

//  Action Request to remove a Maintained Molecule Type Specifier

{
    public:
        Remove_Maint_MolType_Spec_Req(Remove_Maint_MolType_Spec_ReqRqstr*  Requestor_Ptr);
        string Execute(Action_Request*& Repeat_Act_Req_Ptr); 
        int    Get_Number() const;
        string Get_Type() const;
        string Get_Subject() const;
        void  Get_Params(list<string>& Param_List) const;
        
    private:

};
//----------------------------------------------------------------------------------------------------------------

class   Remove_Transformation_Req: public Action_Request

//  Action Request to remove a Transformation

{
    public:
        Remove_Transformation_Req(Remove_Transformation_ReqRqstr*  Requestor_Ptr);
        string Execute(Action_Request*& Repeat_Act_Req_Ptr); 
        int    Get_Number();        
        string Get_Type();
        string Get_Subject();
        void  Get_Params(list<string>& Param_List) const;
        
    private:
        int     Num_To_Remove;
        string  Reactant_Spec;
        string  Product_Spec;
};
//----------------------------------------------------------------------------------------------------------------

class Set_Reactor_Vol_Req: public Action_Request

//  Action Request to set Reactor_Volume

{
    public:
        Set_Reactor_Vol_Req(Set_Reactor_Vol_ReqRqstr*  Requestor_Ptr);
        string Execute(Action_Request*& Repeat_Act_Req_Ptr); 
        int    Get_Number();
        string Get_Type();
        string Get_Subject();
        void  Get_Params(list<string>& Param_List) const;
        
    private:
        int Reactor_Volume;
};
//-----------------------------------------------------------------------------------------------------------------

/*
class Set_Inflow_Rate_Req: public Action_Request

//  Action Request to set the Reactor Inflow Rate

{
    public:
        Set_Inflow_Rate_Req(int Action_Time, int Inflow_Rate);
        string Execute(Action_Request*& Repeat_Act_Req_Ptr); 
        int    Get_Number();
        string Get_Type();
        string Get_Subject();
        void  Get_Params(list<string>& Param_List) const;
        
    private:
        int Reactor_Inflow_Rate;
};
*/
//-----------------------------------------------------------------------------------------------------------------
class Set_Max_Rand_Rate_Const_Constr_Req: public Action_Request

//  Action Request to set the Maximum Random Rate Constant for Constructions
//  Can be used with the Random Rate Constant Model

{
    public:
        Set_Max_Rand_Rate_Const_Constr_Req(Set_Max_Rand_Rate_Const_Constr_ReqRqstr*  Requestor_Ptr);
        string Execute(Action_Request*& Repeat_Act_Req_Ptr); 
        int    Get_Number();
        string Get_Type();
        string Get_Subject();
        void  Get_Params(list<string>& Param_List) const;
        
    private:
        double Max_Rate_Constant;
};
//-----------------------------------------------------------------------------------------------------------------
class Set_Max_Rand_Rate_Const_Fission_Req: public Action_Request

//  Action Request to set the Maximum Rate Constant for Fissions

{
    public:
        Set_Max_Rand_Rate_Const_Fission_Req(Set_Max_Rand_Rate_Const_Fission_ReqRqstr*  Requestor_Ptr);
        string Execute(Action_Request*& Repeat_Act_Req_Ptr); 
        int    Get_Number();
        string Get_Type();
        string Get_Subject();
        void  Get_Params(list<string>& Param_List) const;
        
    private:
        double Max_Rate_Constant;
};
//-----------------------------------------------------------------------------------------------------------------
class Set_Max_Rate_Const_Transformation_Req: public Action_Request

//  Action Request to set the Maximum Rate Constant for Transformations

{
    public:
        Set_Max_Rate_Const_Transformation_Req(Set_Max_Rate_Const_Transformation_ReqRqstr*  Requestor_Ptr);
        string Execute(Action_Request*& Repeat_Act_Req_Ptr); 
        int    Get_Number();
        string Get_Type();
        string Get_Subject();
        void  Get_Params(list<string>& Param_List) const;
        
    private:
        double Max_Rate_Constant;
};
//-----------------------------------------------------------------------------------------------------------------

class Set_Rate_Const_Model_Req: public Action_Request

//  Action Request to set the model to be used for determining Rate Constants

{
    public:
        Set_Rate_Const_Model_Req(Set_Rate_Const_Model_ReqRqstr*  Requestor_Ptr);
        string Execute(Action_Request*& Repeat_Act_Req_Ptr); 
        int    Get_Number();
        string Get_Type();
        string Get_Subject();
        void   Get_Params(list<string>& Param_List) const;
        
    private:
	string Rate_Const_Model_Spec;
};

//----------------------------------------------------------------------------------------------------------------

class Terminate_Scenario_Req: public Action_Request

//  Action Request to stop the Scaenario in which it appears

{
    public:
        Terminate_Scenario_Req(Terminate_Scenario_ReqRqstr*  Requestor_Ptr);
        string Execute(Action_Request*& Repeat_Act_Req_Ptr); 
        int    Get_Number();
        string Get_Type();
        string Get_Subject();
        void  Get_Params(list<string>& Param_List) const;
        
    private:

};

//----------------------------------------------------------------------------------------------------------------



/*
class Set_Num_Maint_MolTypes_Req: public Action_Request

//  Action Request to set the Number of Maintained Molecule Types

{
    public:
        Set_Num_Maint_MolTypes_Req(int Action_Time, int Num_Maint_MolTypes);
        string Execute(Action_Request*& Repeat_Act_Req_Ptr); 
        int    Get_Number();
        string Get_Type();
        string Get_Subject();
        void  Get_Params(list<string>& Param_List) const;
        
    private:
        int Num_Maint_MolTypes;
};  */

#endif  //Action_Request_H
